import cv2
import numpy as np
import onnxruntime as ort
from typing import List, Tuple
import importlib
import subprocess
import os
from urllib.request import urlretrieve
from urllib.error import URLError


def load_image(image_path: str) -> np.ndarray:
    """
    Load an image from the given path and convert it to a NumPy array.

    Args:
        image_path (str): Path to the image file.

    Returns:
        np.ndarray: The loaded image as a NumPy array.
    """
    image = cv2.imread(image_path)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image

def letterbox(
        im: np.ndarray, 
        new_shape: Tuple[int, int] =(640, 640), 
        color: Tuple[int, int, int]=(114, 114,  114), 
        auto: bool=True, 
        scaleup: bool=True, 
        stride: int=32
        ) -> Tuple[np.ndarray, float, Tuple[float, float]]:

    """
    Resize and pad the image while meeting stride-multiple constraints.

    Args:
        im (np.ndarray): Input image as a NumPy array.
        new_shape (Tuple[int, int]): Desired new shape of the image (height, width).
        color (Tuple[int, int, int]): RGB color value for the padding.
        auto (bool): Flag indicating whether to use the minimum rectangle for padding.
        scaleup (bool): Flag indicating whether to scale up the image if needed.
        stride (int): Stride value for meeting the stride-multiple constraints.

    Returns:
        Tuple[np.ndarray, float, Tuple[float, float]]: Resized and padded image, scale ratio, and padding values.
    """

    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding

    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return im, r, (dw, dh)

def generic_yolo_preprocessing(
        inputs: List[str]
        ) -> Tuple[np.ndarray, List[Tuple[np.ndarray, float, Tuple[float, float]]], List[np.ndarray]]:
    """
    Perform generic YOLO preprocessing on a list of input images.

    Args:
        inputs (List[str]): List of input image paths.

    Returns:
        Tuple[np.ndarray, List[Tuple[np.ndarray, float, Tuple[float, float]]], List[np.ndarray]]: Preprocessed data, including the batch of resized images, ratio and padding information for each image, and the original RGB images.
    """
    resize_data = []
    origin_RGB = []
    for image_path in inputs:
        image = load_image(image_path)
        origin_RGB.append(image)
        image, ratio, dwdh = letterbox(image, auto=False)
        image = image.transpose((2, 0, 1))
        image = np.expand_dims(image, 0)
        image = np.ascontiguousarray(image)
        image = image.astype(np.float32)
        image /= 255
        resize_data.append((image, ratio, dwdh))
    np_batch = np.concatenate([data[0] for data in resize_data])
    return np_batch, resize_data, origin_RGB

def generic_ort_inference(
        model_path: str, 
        inputs: np.ndarray, 
        cuda: bool = False
        ) -> np.ndarray:
    """
    Perform inference using the ONNX model with ONNX Runtime.

    Args:
        model_path (str): Path to the ONNX model file.
        inputs (np.ndarray): Input data for the model as a NumPy array.
        cuda (bool): Flag indicating whether to use CUDA for inference (default: True).

    Returns:
        np.ndarray: Inference results as a NumPy array.
    """
    providers = ['CUDAExecutionProvider'] if cuda else ['CPUExecutionProvider']
    session = ort.InferenceSession(model_path, providers=providers)
    outname = [i.name for i in session.get_outputs()]
    inname = [i.name for i in session.get_inputs()]
    detections = session.run(outname,{inname[0]: inputs})
    return detections

def check_and_install_library(library_name):
    try:
        importlib.import_module(library_name)
        print(f"{library_name} is already installed.")
    except ImportError:
        print(f"{library_name} is not installed. Installing...")
        subprocess.check_call(['pip', 'install', library_name])
        print(f"{library_name} has been successfully installed.")

def download_model_weights(model_type: str) -> Tuple[str, str]:

    root = "checkpoints"
    sam_model_types = {
        'base': "vit_b",
        'large': "vit_l",
        'huge': "vit_h"
    }

    match(model_type):
        case "base":
            url = "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth"
        case "large":
            url = "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_l_0b3195.pth"
        case "huge":
            url = "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth"
        case _:
            raise ValueError("Incorrect SAM model type.")

    file_name = os.path.basename(url)
    os.makedirs(root, exist_ok=True)
    file_path = os.path.join(root, file_name)
    
    if not os.path.isfile(file_path):
        try:        
            print(f"Downloading SAM weights to {file_path} from {url}")
            urlretrieve(url, file_path)
        except (URLError, IOError) as _:
            print(f"Could not download SAM Model Weights.")
    return sam_model_types[model_type], file_path

def check_yolo_version(yolo_version: str) -> None:
    implemented_yolo_versions = ["6", "7", "8", "nas"]
    if not yolo_version in implemented_yolo_versions:
        raise NotImplementedError(f"YOLO version not implemented.")
    
def check_device(device:str) -> None:
    if not device in ["cuda", "cpu"]:
        raise ValueError("Please specify 'cpu' or 'cuda' for device.")
    
def check_sam_source(sam_source: str) -> None:
    if not sam_source in ["Meta", "HuggingFace"]:
        raise NotImplementedError(f"SAM source not implemented.")
    
def check_sam_model_type(sam_model_type: str) -> None:
    if not sam_model_type in ["base", "large", "huge"]:
        raise ValueError("Incorrect SAM model type.")
    
def perform_samyol_input_checks(
    yolo_version: str,
    device: str,
    sam_source: str,
    sam_model_type: str
) -> None:
    check_yolo_version(yolo_version)
    check_device(device)
    check_sam_source(sam_source)
    check_sam_model_type(sam_model_type)



from .utils import check_and_install_library, download_model_weights
from typing import List, Dict
import numpy as np
import torch


class SAM:
    @staticmethod
    def predict_from_Meta(
        sam_model_type: str,
        original_RGB: List[np.ndarray], 
        obj_det_predictions: List[Dict], 
        device: str
    ) -> List[Dict]:
        """
        Predict object segmentation using SAM from Meta.

        Args:
            sam_model_type (str): SAM model type to use: base, large or huge.
            original_RGB (List[np.ndarray]): List of RGB images.
            obj_det_predictions (List[Dict]): Object detection predictions.
            device (str): Device for inference.

        Returns:
            List[Dict]: List of object segmentation predictions.
        """
        # Load Model
        check_and_install_library('segment_anything')
        from segment_anything import SamPredictor, sam_model_registry
        model_type, model_path = download_model_weights(sam_model_type)
        sam = sam_model_registry[model_type](checkpoint=model_path).to(device)
        predictor = SamPredictor(sam)

        # Get predictions from SAM
        object_segmentation_predictions = []

        image_ids = list(set([d['image_id'] for d in obj_det_predictions]))

        for image_id in image_ids:
            # Filter the data based on the current image_id
            filtered_data = [d for d in obj_det_predictions if d['image_id'] == image_id]

            # Extract the bounding boxes for the current image_id
            bboxes = [d['bbox'] for d in filtered_data]

            # Extract the class ids and class labels for the current image_id
            class_ids = [d['class_id'] for d in filtered_data]
            class_labels = [d['class_label'] for d in filtered_data]

            # Get the image based on the current image_id
            image = original_RGB[image_id]

            # Transform input BBoxes
            input_boxes = torch.tensor(
                bboxes, 
                device=predictor.device
            )

            predictor.set_image(image)
            transformed_boxes = predictor.transform.apply_boxes_torch(input_boxes, image.shape[:2])
            masks, scores, _ = predictor.predict_torch(
                point_coords=None,
                point_labels=None,
                boxes=transformed_boxes,
                multimask_output=True
            )
            idx_max_iou = torch.argmax(scores.view(-1, 3), dim=1).tolist()

            object_segmentation_predictions.append({
                'image_id': image_id,
                'class_id': class_ids,
                'class_label': class_labels,
                'score': [scores.squeeze()[i, j, ...].item() for i, j in enumerate(idx_max_iou)],
                'bbox': bboxes,
                'masks': [masks[i, j, ...].cpu().numpy() for i, j in enumerate(idx_max_iou)]
            })

        return object_segmentation_predictions

    @staticmethod
    def predict_from_HuggingFace(
        sam_model_type: str,
        original_RGB: List[np.ndarray], 
        obj_det_predictions: List[Dict], 
        device: str
    ) -> List[Dict]:
        """
        Predict object segmentation using SAM from Meta.

        Args:
            sam_model_type (str): SAM model type to use: base, large or huge.
            original_RGB (List[np.ndarray]): List of RGB images.
            obj_det_predictions (List[Dict]): Object detection predictions.
            device (str): Device for inference.

        Returns:
            List[Dict]: List of object segmentation predictions.
        """
        # Load SAM
        check_and_install_library('transformers')
        from transformers import SamModel, SamProcessor

        sam_model = SamModel.from_pretrained(f"facebook/sam-vit-{sam_model_type}").to(device)
        sam_processor = SamProcessor.from_pretrained(f"facebook/sam-vit-{sam_model_type}")

        # Get predictions from SAM
        object_segmentation_predictions = []

        image_ids = list(set([d['image_id'] for d in obj_det_predictions]))

        for image_id in image_ids:
            # Filter the data based on the current image_id
            filtered_data = [d for d in obj_det_predictions if d['image_id'] == image_id]

            # Extract the bounding boxes for the current image_id
            bboxes = [d['bbox'] for d in filtered_data]

            # Extract the class ids and class labels for the current image_id
            class_ids = [d['class_id'] for d in filtered_data]
            class_labels = [d['class_label'] for d in filtered_data]

            # Get the image based on the current image_id
            image = original_RGB[image_id]

            # Perform the inference for the current image_id
            inputs = sam_processor(image, input_boxes=[[bboxes]], return_tensors="pt").to(device)
            image_embeddings = sam_model.get_image_embeddings(inputs["pixel_values"])
            inputs.pop("pixel_values", None)
            inputs.update({"image_embeddings": image_embeddings})
            with torch.no_grad():
                outputs = sam_model(**inputs)
            masks = sam_processor.image_processor.post_process_masks(outputs.pred_masks.cpu(), inputs["original_sizes"].cpu(), inputs["reshaped_input_sizes"].cpu())
            scores = outputs.iou_scores

            # Reshape the tensor to have size (N, 3)
            reshaped_iou_scores = outputs.iou_scores.squeeze()
            idx_max_iou = torch.argmax(reshaped_iou_scores.view(-1, 3), dim=1).tolist()

            object_segmentation_predictions.append({
                'image_id': image_id,
                'class_id': class_ids,
                'class_label': class_labels,
                'score': [outputs.iou_scores.squeeze()[i, j, ...].item() for i, j in enumerate(idx_max_iou)],
                'bbox': bboxes,
                'masks': [masks[0][i, j, ...].cpu().numpy() for i, j in enumerate(idx_max_iou)]
            })

        return object_segmentation_predictions


from samyol.yolo_preprocessing import YOLOPreProcessing
from samyol.yolo_inference import YOLOInference
from samyol.yolo_postprocessing import YOLOPostProcessing
from samyol.prediction_results import SAMYOLPredictions
from samyol.sam_inference import SAM
from samyol.utils import perform_samyol_input_checks
from typing import Union, List, Optional, Dict, Callable

class SAMYOL:   
    def __init__(
        self,
        yolo_model_path: str,
        yolo_version: str,
        sam_model_type: str,
        sam_source: str,
        class_labels: List[str],
        device: str,
        extra_args: Optional[Dict] = None,
        ):
        """
        Initialize the SAMYOL object.

        Args:
            yolo_model_path (str): Path to the YOLO model.
            yolo_version (str): Version of the YOLO model to use.
            sam_model_type (str): SAM model type to use: base, large or huge.
            sam_source (str): SAM source: HuggingFace or facebook repo.
            class_labels (List[str]): List of class labels.
            device (str): Device to use for inference.
            extra_args (Dict, optional): Extra arguments to be passed to the YOLO-NAS inference step. Defaults to None.
        """
        perform_samyol_input_checks(
            yolo_version=yolo_version,
            sam_source=sam_source,
            sam_model_type=sam_model_type,
            device=device
        )
        self.yolo_model_path = yolo_model_path
        self.yolo_version = yolo_version
        self.sam_model_type = sam_model_type
        self.sam_source = sam_source
        self.class_labels = class_labels
        self.device = device
        self.kwargs = extra_args if extra_args is not None else {}

    def predict(
            self,
            input_paths: Union[str, List[str]],
    ) -> SAMYOLPredictions:
        """
        Predicts object segmentation using the SAMYOL model.

        Args:
            input_paths (Union[str, List[str]]): Path(s) to the input image(s).

        Returns:
            SAMYOLPredictions: Object containing the input images and segmentation predictions.
        """
        if not isinstance(input_paths, List):
            input_paths = [input_paths]

        yolo_pipeline = self.get_yolo_pipeline(self.yolo_version)
        preprocessed_data = yolo_pipeline['preprocessing'](input_paths)
        outputs = yolo_pipeline['inference'](self.yolo_model_path, preprocessed_data, self.device, **self.kwargs)
        obj_det_predictions = yolo_pipeline['postprocessing'](outputs, self.class_labels)
        
        sam_pipeline = self.get_sam_pipeline(self.sam_source)
        object_segmentation_predictions = sam_pipeline['sam_inference'](self.sam_model_type, preprocessed_data[-1], obj_det_predictions, self.device)

        return SAMYOLPredictions(
            images=preprocessed_data[-1], 
            predictions=object_segmentation_predictions,
            )
    

    @staticmethod
    def get_yolo_pipeline(version: str) -> Dict[str, Callable]:
        """
        Get the YOLO pipeline components based on the specified version.

        Args:
            version (str): Version of the YOLO model.

        Returns:
            Dict[str, Callable]: Dictionary containing the YOLO pipeline components.
        """
        run_yolo_preprocessing = getattr(YOLOPreProcessing, f"get_yolo_{version}_preprocessing")
        run_yolo_inference = getattr(YOLOInference, f"get_yolo_{version}_inference")
        run_yolo_postprocessing = getattr(YOLOPostProcessing, f"get_yolo_{version}_postprocessing")
        return {
            'preprocessing': run_yolo_preprocessing, 
            'inference': run_yolo_inference, 
            'postprocessing': run_yolo_postprocessing
        }

    @staticmethod
    def get_sam_pipeline(source: str) -> Dict[str, Callable]:
        """
        Get the SAM pipeline components based on the specified source.

        Args:
            source (str): Source for SAM model inference.

        Returns:
            Dict[str, Callable]: Dictionary containing the SAM pipeline components.
        """
        run_sam_inference = getattr(SAM, f"predict_from_{source}")
        return {
            'sam_inference': run_sam_inference
                }



from .utils import generic_yolo_preprocessing, load_image
from typing import List, Tuple
import numpy as np


class YOLOPreProcessing():
    
    @staticmethod
    def get_yolo_6_preprocessing(
        inputs: List[str]
        ) -> Tuple[np.ndarray, List[Tuple[np.ndarray, float, Tuple[float, float]]], List[np.ndarray]]:
        """
        Perform YOLO-6 preprocessing on the inputs.

        Args:
            inputs (List[str]): List of input image paths.

        Returns:
            Tuple[np.ndarray, List[Tuple[np.ndarray, float, Tuple[float, float]]], List[np.ndarray]]: Preprocessed data, including the batch of resized images, ratio and padding information for each image, and the original RGB images.
        """
        return generic_yolo_preprocessing(inputs)

    @staticmethod
    def get_yolo_7_preprocessing(
        inputs: List[str]
        ) -> Tuple[np.ndarray, List[Tuple[np.ndarray, float, Tuple[float, float]]], List[np.ndarray]]: 
        """
        Perform YOLO-7 preprocessing on the inputs.

        Args:
            inputs (List[str]): List of input image paths.

        Returns:
            Tuple[np.ndarray, List[Tuple[np.ndarray, float, Tuple[float, float]]], List[np.ndarray]]: Preprocessed data, including the batch of resized images, ratio and padding information for each image, and the original RGB images.
        """
        return generic_yolo_preprocessing(inputs)

    @staticmethod
    def get_yolo_8_preprocessing(
        inputs: List[str]
        ) -> List[np.ndarray]:
        """
        Perform YOLO-NAS preprocessing on the inputs.

        Args:
            inputs List[str]: List of input image path(s) .

        Returns:
            inputs List[np.ndarray]: List of input images.
        """
        origin_RGB = [load_image(image_path) for image_path in inputs]

        return [origin_RGB]
    
    @staticmethod
    def get_yolo_nas_preprocessing(
        inputs: List[str]
        ) -> List[np.ndarray]:
        """
        Perform YOLO-NAS preprocessing on the inputs.

        Args:
            inputs List[str]: List of input image paths .

        Returns:
            inputs List[np.ndarray]: List of input images.
        """
        origin_RGB = [load_image(image_path) for image_path in inputs]

        return [inputs, origin_RGB]


import torch
from typing import List, Tuple
from pathlib import Path

from .utils import generic_ort_inference, check_and_install_library


class YOLOInference():

    @staticmethod
    def get_yolo_6_inference(model_path: str, inputs: Tuple, device: str) -> Tuple:
        """
        Perform YOLO-6 inference.

        Args:
            model_path (str): Path to the YOLOv6 model.
            inputs (Tuple): Input data.
            device (str): Device to use for inference.

        Returns:
            Tuple: Inference results, resize data, and original RGB images.
        """
        np_batch, resize_data, origin_RGB = inputs
        cuda = True if device == "cuda" else False
        detections = generic_ort_inference(model_path, np_batch, cuda=cuda)
        return detections, resize_data, origin_RGB
    
    @staticmethod
    def get_yolo_7_inference(model_path: str, inputs: Tuple, device:str) -> Tuple:
        """
        Perform YOLO-7 inference.

        Args:
            model_path (str): Path to the YOLOv7 model.
            inputs (Tuple): Input data.
            device (str): Device to use for inference.

        Returns:
            Tuple: Inference results, resize data, and original RGB images.
        """
        np_batch, resize_data, origin_RGB = inputs
        cuda = True if device == "cuda" else False
        detections = generic_ort_inference(model_path, np_batch, cuda=cuda)
        return detections, resize_data, origin_RGB

    @staticmethod
    def get_yolo_8_inference(model_path: str, inputs: List, device:str) -> List:
        """
        Perform YOLO-8 inference.

        Args:
            model_path (str): Path to the YOLOv8 model.
            inputs (Tuple): Input data.
            device (str): Device to use for inference.

        Returns:
            Tuple: Inference results, resize data, and original RGB images.
        """
        check_and_install_library('ultralytics')
        from ultralytics import YOLO
        model = YOLO(model_path)
        detections = [model.predict(
            image, 
            conf=0.35, 
            iou=0.65, 
            verbose=False, 
            device=str(torch.cuda.current_device()) if device=="cuda" else "cpu"
        ) for image in inputs[0]]
        return detections

    
    @staticmethod
    def get_yolo_nas_inference(
        model_path: str, 
        inputs: List,
        device:str,
        model_type: str, 
        num_classes: int
        ) -> List:
        """
        Perform YOLO-NAS inference.

        Args:
            model_path (str): Path to the YOLO-NAS model.
            inputs (List): Input data.
            device (str): Device to use for inference.
            model_type (str): Model type.
            num_classes (int): Number of classes.

        Returns:
            List: List of detection results.
        """
        check_and_install_library('super_gradients')
        import super_gradients as sg
        from super_gradients.training import models
        sg.setup_device(device=device)
        model = models.get(
            model_type,
            num_classes=num_classes,
            checkpoint_path=model_path
        ).to(torch.device(device))
        detections = model.predict(inputs[0])
        return detections


from setuptools import setup, find_packages

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
  name="SAMYOL",
  version="1.0.3",
  description="Combines YOLO models and SAM",
  packages=find_packages(),
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/Jawher-Ben-Abdallah/SAMYOL",
  author="Jawher Ben Abdallah - Rim Sleimi",
  author_email="jawher.b.abdallah@gmail.com",
  license='Apache License 2.0',
  classifiers=[
      "Programming Language :: Python :: 3.10",
    ],
    keywords="YOLOv6, YOLOv7, YOLOv8, YOLO-NAS SAM, Object_Detection, Segmentation",
    install_requires=[
        "matplotlib>=3.2.2",
        "numpy>=1.21.6",
        "opencv-python>=4.6.0",
        "Pillow>=7.1.2",
        "PyYAML>=5.3.1",
        "requests>=2.23.0",
        "scipy>=1.4.1",
        "torch>=1.7.0",
        "torchvision>=0.8.1",
        "tqdm>=4.64.0", 
        "onnxruntime"
    ],
    
    extras_require={"dev" : "twine>=4.0.2"},
    python_requires=">=3.10",
)


from typing import List, Dict
import numpy as np
import cv2
import random
import matplotlib.pyplot as plt
from PIL import Image
import json


class SAMYOLPredictions():
    def __init__(
            self,
            images: List[np.ndarray],
            predictions: List[Dict]
    ):
        """
        Initialize SAMYOLPredictions.

        Args:
            images (List[np.ndarray]): List of input images.
            predictions (List[Dict]): List of prediction dictionaries.
        """
        self.images = images
        self.predictions = predictions

    def display(self, index: int) -> None:
        """
        Display the bounding boxes and masks for a specific image.

        Args:
            index (int): Index of the image to display.
        """
        image = self.images[index]
        target_predictions = self.predictions[index]

        # Create a subplot for displaying the image, bounding boxes, and masks
        fig, ax = plt.subplots(figsize=(6, 6))

        # Plot the image
        ax.imshow(image)

        # Plot the masks with low opacity
        colors = []
        for mask, class_id in zip(target_predictions['masks'], target_predictions['class_id']):
            color = np.concatenate([np.random.random(3), np.array([0.6])], axis=0)
            h, w = mask.shape[-2:]
            bbox_mask = mask.reshape(h, w, 1) * color.reshape(1, 1, -1)
            ax.imshow(bbox_mask)

        # Plot the bounding boxes
        for bbox, class_label in zip(target_predictions['bbox'], target_predictions['class_label']):
            x1, y1, x2, y2 = bbox
            color = (random.random(), random.random(), random.random()) # Generate a random color for each class_label
            rect = plt.Rectangle((x1, y1), x2 - x1, y2 - y1, fill=False, edgecolor=color, linewidth=2)
            ax.add_patch(rect)
            ax.text(x1, y1 - 5, class_label, fontsize=8, color="black", weight="bold", bbox=dict(facecolor=color, edgecolor="black", alpha=0.8, pad=0.8))


        # Configure the plot
        ax.axis('off')
        plt.tight_layout()

        # Show the plot
        plt.show()

    def save(self, 
             save_dir: str="./", 
             filename: str="default", 
             fuse_masks: bool=False, 
             save_metadata: bool=False, 
             image_id:int=0, 
             mask_format: str="png") -> None:

        """
        Save the masks as images and metadata as JSON file.

        Args:
            save_dir (str): Directory to save the images.
            filename (str): Filename for the saved images.
            fuse_masks (bool): Whether to merge all masks or save them separately.
            save_metadata (bool): Whether to save metadata as a JSON file.
            image_id (int): Index of the image to save.
            mask_format (str): Image format to save.
        """

        filtered_data = self.predictions[image_id]
        masks = filtered_data['masks']
        if fuse_masks:
            # Merge all masks
            masks = np.logical_or.reduce(np.array(masks))
            cv2.imwrite(f"{save_dir}/{filename}.{mask_format}", masks * 255)
        else:
            # Per Mask
            class_ids = self.predictions[image_id]['class_id']
            masks = [mask * (class_id + 1) for (mask, class_id) in zip (masks, class_ids)]
            masks = np.stack(masks, axis=2)
            masks = masks.astype(np.uint8)
            pil_masks = Image.fromarray(masks)
            pil_masks.save(f"{save_dir}/{filename}.{mask_format}")
        
        if save_metadata:
            metadata = {
                'image_id': image_id, 
                'class_id': filtered_data['class_id'], 
                'class_label':filtered_data['class_label'],
                'score':  filtered_data['score'], 
                'bbox': filtered_data['bbox'],
                }
            # Save dictionary as JSON file
            with open(f"{save_dir}/{filename}.json", "w", encoding="utf-8") as json_file:
                json.dump(metadata, json_file, indent=4)



import numpy as np
from typing import List

class YOLOPostProcessing():

    @staticmethod
    def get_yolo_6_postprocessing(detections:tuple, class_labels: List[str]) -> List[dict]:
        """
        Perform YOLOv6 post-processing on the detections.

        Args:
            detections (tuple): Output data from the YOLOv6 model.
            class_labels (List[str]): List of class labels.

        Returns:
            List[dict]: List of object detection predictions.
        """
        object_detection_predictions = []
        detections, rezise_data, origin_RGB = detections
        for i in range(detections[0].shape[0]):
            obj_num = detections[0][i]
            boxes = detections[1][i]
            scores = detections[2][i]
            cls_id = detections[3][i]
            image = origin_RGB[i]
            img_h, img_w = image.shape[:2]
            ratio, dwdh = rezise_data[i][1:]
            for num in range(obj_num[0]):
                box = boxes[num]
                score = round(float(scores[num]), 3)
                class_id = int(cls_id[num])
                box -= np.array(dwdh*2)
                box /= ratio
                box = box.round().astype(np.int32).tolist()
                x1 = max(0, box[0])
                y1 = max(0, box[1])
                x2 = min(img_w, box[2])
                y2 = min(img_h, box[3])
                object_detection_predictions.append({
                    'image_id': i,
                    'class_label': class_labels[class_id],
                    'class_id': class_id,
                    'score': score,
                    'bbox': [x1, y1, x2, y2]
                })
        return object_detection_predictions


    @staticmethod
    def get_yolo_7_postprocessing(detections: tuple, class_labels: List[str]) -> List[dict]:
        """
        Perform YOLOv7 post-processing on the detections.

        Args:
            detections (tuple): Output data from the YOLOv7 model.
            class_labels (List[str]): List of class labels.

        Returns:
            List[dict]: List of object detection predictions.
        """
        object_detection_predictions = []
        detections, rezise_data, _ = detections
        for batch_id, x0, y0, x1, y1, cls_id, score in detections[0]:
            ratio, dwdh = rezise_data[int(batch_id)][1:]
            box = np.array([x0, y0, x1, y1])
            box -= np.array(dwdh*2)
            box /= ratio
            box = box.round().astype(np.int32).tolist()
            cls_id = int(cls_id)
            score = round(float(score),3)
            object_detection_predictions.append({
                    'image_id': int(batch_id),
                    'class_label': class_labels[cls_id],
                    'class_id': cls_id,
                    'score': score,
                    'bbox': box
                })
        return object_detection_predictions
    
    @staticmethod
    def get_yolo_8_postprocessing(detections: List[object], class_labels: List[str]) -> List[dict]:
        """
        Perform YOLOv8 post-processing on the detections.

        Args:
            detections (List[object]): Detections from the YOLOv8 model.
            class_labels (List[str]): List of class labels.

        Returns:
            List[dict]: List of object detection predictions.
        """
        object_detection_predictions = []
        for i, detection in enumerate(detections):
            boxes = detection[0].boxes
            for box in boxes:
                class_id = int(box.cls.item())
                object_detection_predictions.append(
                    {
                        'image_id': i,
                        'class_label': class_labels[class_id],
                        'class_id': class_id,
                        'score': box.conf.item(),
                        'bbox': box.xyxy[0].cpu().numpy().round().astype(np.int32).tolist()
                    }
                )
        return object_detection_predictions


    @staticmethod
    def get_yolo_nas_postprocessing(detections: List[object], class_labels: List[str]) -> List[dict]:
        """
        Perform YOLO-NAS post-processing on the detections.

        Args:
            detections (List[object]): Detections from the YOLO-NAS model.
            class_labels (List[str]): List of class labels.

        Returns:
            List[dict]: List of object detection predictions.
        """
        object_detection_predictions = []
        for i, detection in enumerate(detections):
            labels = detection.prediction.labels
            confidence = detection.prediction.confidence
            bboxes = detection.prediction.bboxes_xyxy

            for label, conf, bbox in zip(labels, confidence, bboxes):
                object_detection_predictions.append({
                    'image_id': i,
                    'class_label': class_labels[int(label)],
                    'class_id': int(label),
                    'confidence': conf,
                    'bbox': bbox.round().astype(np.int32).tolist()
                })
        return object_detection_predictions